/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

#define ADDR		0x84
#define NMEA_LEN	83
#define READ_LEN	100

static uint8_t gotNewData;
static gps_t gpsInfo;
static uint8_t startupDelay; // Powered on, waiting a few milliseconds before sending UBX commands
static millis_t timer;
static uint8_t useUART;
static char sentence[NMEA_LEN];
static uint8_t idx;

static void ubxChecksum(uint8_t* data, uint8_t len)
{
	uint8_t CK_A = 0;
	uint8_t CK_B = 0;
	for(uint8_t i=2;i<len-2;i++)
	{
		CK_A = CK_A + data[i];
		CK_B = CK_B + CK_A;
	}
	data[len - 2] = CK_A;
	data[len - 1] = CK_B;
}

static uint8_t checksum(char* sentence)
{
	uint8_t checksum = 0;
	while(*sentence)
	{
		if(*sentence == '*')
		{
			uint8_t c = hex2dec(sentence + 1);
			return c == checksum;
		}
		checksum ^= *sentence++;
	}
	return 0;
}

static void strncpyNullTerm(char* dst, char* src, uint8_t num)
{
	strncpy(dst, src, num);
	dst[num - 1] = '\0';
}

static void updateLedState(void)
{
	if(gpsInfo.state == GPS_STATE_OFF)
		led_off(0);
	else if(gpsInfo.state == GPS_STATE_LOCKING)
		led_set(0, 100, 100); // Fast flash for locking
	else
		led_set(0, 100, 1000); // Slow flash for locked
}

static void processSentence(void)
{
	if(!checksum(sentence + 1))
	{
		GPS_DBG("Checksum failed\n");
		return;
	}

	GPS_DBG("");

	uint8_t index = 0;

	char* ptrSentence = sentence;
	char* field;
	while((field = getNextField(&ptrSentence)) != NULL)
	{
		// FAKE (4 digits for the decimal minutes fraction instead of 5 and 3 digits after time instead of 2):
		// $GPRMC,205052.000,A,xxxx.xxxx,N,xxxxx.xxxx,E,2.15,350.37,080218,,,A,AB,
		// REAL:
		// $GPRMC,213620.00,A,xxxx.xxxxx,N,xxxxx.xxxxx,E,0.697,,080218,,,A,AB,

		if(index == 0)
		{
			if(strcmp(field, "$GPRMC") != 0) // Only process RMC sentences
			{
				GPS_DBG_RAW("Not GPRMC (%s)\n", field);
				return;
			}
		}
		else if(index == 1)
			strncpyNullTerm(gpsInfo.time, field, sizeof(gpsInfo.time));
		else if(index == 2)
			gpsInfo.state = (field[0] == 'A') ? GPS_STATE_LOCKED : GPS_STATE_LOCKING;
		else if(index == 3)
			strncpyNullTerm(gpsInfo.latitude, field, sizeof(gpsInfo.latitude));
		else if(index == 4)
			gpsInfo.latitudeHemisphere = field[0];
		else if(index == 5)
			strncpyNullTerm(gpsInfo.longitude, field, sizeof(gpsInfo.latitude));
		else if(index == 6)
			gpsInfo.longitudeMeridian = field[0];
		else if(index == 7)
			gpsInfo.speed = (atoi(field) * 185) / 100; // Knots to KPH (knots * 1.852 = KPH)
			//gpsInfo.speed = (atoi(field) * 115) / 100; // Knots to MPH (knots * 1.15 = MPH)
		else if(index == 9)
			strncpyNullTerm(gpsInfo.date, field, sizeof(gpsInfo.date));

		GPS_DBG_RAW("%s,", field);

		index++;
	}

	if(!strlen(gpsInfo.time))
		strcpy(gpsInfo.time, "000000");
	if(!strlen(gpsInfo.date))
		strcpy(gpsInfo.date, "000000");

	GPS_DBG_RAW("\n");

//	if(strlen(gpsInfo.time))
//		GSM_DBG("GPS has time\n");
//	if(strlen(gpsInfo.date))
//		GSM_DBG("GPS has date\n");

	gotNewData = 1;

	if(gpsInfo.state == GPS_STATE_LOCKED)
	{
		//  NMEA Degrees and decimal minutes (DMM) to Decimal degrees (DD)
		uint32_t tmp;
		
		// Some GPS modules (mainly Chinese fakes) only output 4 digits for the decimal minutes fraction instead of 5, so for those we multiply the values by 10

		gpsInfo.deg_lat_frac = atoi(gpsInfo.latitude + 2) * 100000;
		tmp = atoi(gpsInfo.latitude + 5);
		gpsInfo.deg_lat_frac += (strlen(gpsInfo.latitude + 5) == 4) ? tmp * 10 : tmp;
		gpsInfo.deg_lat_frac /= 60;

		gpsInfo.deg_lon_frac = atoi(gpsInfo.longitude + 3) * 100000;
		tmp = atoi(gpsInfo.longitude + 6);
		gpsInfo.deg_lon_frac += (strlen(gpsInfo.longitude + 6) == 4) ? tmp * 10 : tmp;
		gpsInfo.deg_lon_frac /= 60;

		gpsInfo.deg_lat = atoi(gpsInfo.latitude) / 100;
		gpsInfo.deg_lon = atoi(gpsInfo.longitude) / 100;

		if(gpsInfo.latitudeHemisphere != 'N')
		{
			if(gpsInfo.deg_lat == 0)
				gpsInfo.deg_lat_isNegZero = 1;
			else
				gpsInfo.deg_lat *= -1;
		}
		else
			gpsInfo.deg_lat_isNegZero = 0;

		if(gpsInfo.longitudeMeridian != 'E')
		{
			if(gpsInfo.deg_lon == 0)
				gpsInfo.deg_lon_isNegZero = 1;
			else
				gpsInfo.deg_lon *= -1;
		}
		else
			gpsInfo.deg_lon_isNegZero = 0;

		gpsInfo.lastLock = millis();
	}
	else
	{
		gpsInfo.deg_lat_isNegZero = 0;
		gpsInfo.deg_lat = 0;
		gpsInfo.deg_lat_frac = 0;
		gpsInfo.deg_lon_isNegZero = 0;
		gpsInfo.deg_lon = 0;
		gpsInfo.deg_lon_frac = 0;
	}

	updateLedState();

	//GSM_DBG("%s %s %s\n", gpsInfo.latitude, gpsInfo.longitude, gpsInfo.state == GPS_STATE_LOCKED ? "OK" : "WAITING");
}

static void getNewData(uint8_t total)
{
	while(total)
	{
		uint8_t len = (total > READ_LEN) ? READ_LEN : total;

		// If the buffer is full then just reset it
		if(idx >= NMEA_LEN)
			idx = 0;

		// Make sure we don't write too much to the buffer
		if(idx + len > NMEA_LEN)
			len = NMEA_LEN - idx;

		total -= len;

		if(!useUART)
		{
			if(i2c_read(ADDR, sentence + idx, len) != len)
			{
				GPS_DBG("I2C Read error\n");
				return;
			}
		}
		else
		{
			if(uartsoft_read(sentence + idx, len) != len)
			{
				GPS_DBG("UARTSOFT Read error\n");
				return;
			}
		}

		// TODO what about 0xFF?

		// Look for end of sentence in new data
		for(uint8_t i=0;i<len;i++)
		{
			if(sentence[idx] != '\n')
				idx++;
			else // End found
			{
				if(sentence[0] == '$' && sentence[idx - 1] == '\r')
				{
					sentence[idx - 1] = '\0';
					processSentence();
				}
				memcpy(sentence, sentence + idx + 1, len - i - 1); // Move unprocessed data down to the beginning
				idx = 0;
			}
		}
	}
}

static void gpsStuff_i2c(void)
{
	uint8_t buff[2];

	buff[0] = 253;
	if(i2c_write(ADDR, buff, 1, 1) != 1) // Set start address
	{
		GPS_DBG("I2C Write error\n");
		return;
	}
	
	if(i2c_read(ADDR, buff, 2) != 2) // Read 2 bytes to get length of data available
	{
		GPS_DBG("I2C Read error\n");
		return;
	}
	uint16_t total = buff[0]<<8 | buff[1];

	//if(total)
	//	GPS_DBG("Data len: %u\n", total);

	getNewData(total);
}

static void gpsStuff_uartsoft(void)
{
	uint8_t total = uartsoft_available();
	getNewData(total);
}

void gps_init()
{
	// A9 = SCL
	// A10 = SDA
	// B1 = Power gate
	// B3 = Batt backup

	GPIOB->ODR |= GPIO1_OUT_HIGH;
	GPIOB->MODER |= GPIO3_MODE_OUTPUT | GPIO1_MODE_OUTPUT;

	uartsoft_disable();
}

gps_t* gps_get(uint8_t* isNewData)
{
	if(isNewData != NULL)
	{
		*isNewData = gotNewData;
		gotNewData = 0;
	}

	return &gpsInfo;
}

void gps_on()
{
	if(gpsInfo.state == GPS_STATE_OFF)
	{
		GPS_DBG("** ON **\n");

		GPIOB->ODR &= GPIO1_OUT_MASK;
		uartsoft_enable();
		i2c_gps(I2C_GPS_ON);
		timer = millis();
		startupDelay = 1;
		useUART = 0;
		gpsInfo.state = GPS_STATE_LOCKING;
		gpsInfo.lastLock = millis(); // TODO 1 hour in the past?
		updateLedState();
	}
}

void gps_off()
{
	i2c_gps(I2C_GPS_OFF);
	uartsoft_disable();
	GPIOB->ODR |= GPIO1_OUT_HIGH;
	gpsInfo.state = GPS_STATE_OFF;
}

void gps_update()
{
	if(gpsInfo.state == GPS_STATE_OFF)
	{
		updateLedState();
		return;
	}
	else if(startupDelay)
	{
		// TODO look for message ready from I2C instead?
		if(millis() - timer >= 1000)
		{
			startupDelay = 0;
			timer = millis();
			
			// Turn off unwanted messages
			
			i2c_gps(I2C_GPS_DATA);
			
			uint8_t req[] = {
				0xB5,
				0x62,
				0x06,
				0x01,
				0x08, // length
				0x00, // length
				0xF0, // message class
				0x00, //0x01, // message id
				0x00,
				0x01,
				0x01,
				0x01,
				0x01,
				0x01,
				0x00, //0x05, // checksum
				0x00 //0x39, // checksum
			};

			GPS_DBG("I2C config...\n");
			for(uint8_t i=0;i<=5;i++)
			{
				if(i != 4) // 4 is GPRMC
				{
					req[7] = i;
					ubxChecksum(req, sizeof(req));
					if(i2c_write(ADDR, req, sizeof(req), 1) == 0)
					{
						GPS_DBG("I2C Failed, using UART instead\n");
						useUART = 1;
						break;
					}
				}
			}
			GPS_DBG("Config complete\n");
			
			// GPS off test
			//uint8_t GPSoff[] = {0xB5, 0x62, 0x02, 0x41, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x4D, 0x3B};
			//i2c_write(ADDR, GPSoff, sizeof(GPSoff), 1);

			i2c_gps(I2C_GPS_ON);
		}
	}
	else if(useUART)
		gpsStuff_uartsoft();
	else if(millis() - timer >= 500)
	{
		timer = millis();
		i2c_gps(I2C_GPS_DATA);
		gpsStuff_i2c();
		i2c_gps(I2C_GPS_ON);
	}
}
