/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

#define BUFFLEN_LINE 255 // Must be large enough to hold an entire SMS PDU (255 is not enough for a 160 character SMS, but it's enough for the commands that we expect which are only around 15 characters. Would probs need to be 400-500 bytes for an entire PDU)

#define STARTUP_SEQ_NONE		0
#define STARTUP_SEQ_POWERON		1
#define STARTUP_SEQ_CMDDELAY	2
#define STARTUP_SEQ_SMSREADY	3
#define STARTUP_SEQ_OFFSET		4 // Must be last

#define CMD_DELAY				500 // How long to wait after turning power save off before sending a command
#define PWR_SAVE_DELAY			2000 // How long to wait after turning power save off (to send a command) before turning it back on again
#define TURN_BACK_ON_DELAY		4000 // How long to wait after powering off before being able to power on again (need to wait for the capacitor to discharge)
#define TURN_ON_CMD_DELAY		500 // How long to wait after powering on before sending startup commands
#define SMSREADY_TIMEOUT		10000 // How long to wait for "SMS Ready" at startup

#define CMD_TIMEOUT_RESET_COUNT	5

typedef struct {
	const char* str;
	uint32_t timeout;
	uint8_t maxRetries; // Retries on error, not timeout
	uint8_t retryDelay;
} cmd_startup_t;

typedef struct {
	gsmCmdState_t state;
	char command[GSM_CMD_MAX_LEN];
	millis_t timeout;
	millis_t time;
	//millis_t lastTry;
	//uint8_t tries;
	gsmcmd_callback_t callback;
} cmd_t;

static const cmd_startup_t cmd_at = {"AT\r", 250, 25, 50};
static const cmd_startup_t cmd_config1 = {"AT+GSMBUSY=1;+EXUNSOL=\"SQ\",1\r", 5000, 25, 50}; // Reject calls, show signal strength
static const cmd_startup_t cmd_config2 = {"AT+CREG=1;+CSCLK=1\r", 5000, 25, 50}; // Power save / Mode 1 = Manual (DTR pin) / Mode 2 = Auto (after 5 seconds)

static const cmd_startup_t* startupConfig[] = {
	&cmd_at,
	&cmd_config1,
	&cmd_config2,
	NULL
};

static millis_t startupTimer;
static uint8_t startupSequence;
static millis_t lastTurnoffTime;
static millis_t powerSave;
static uint8_t signal;
static uint8_t creg;
static gsm_state_t state;
static cmd_t cmd;
static char lineBuffer[BUFFLEN_LINE];
static uint8_t timeoutCounter;

static millis_t startupTime;

// DTR
#define PWRSAVE_OFF()	(GPIOB->ODR &= GPIO5_OUT_MASK) // uart_send("\r", 1); // For auto mode power save
#define PWRSAVE_ON()	(GPIOB->ODR |= GPIO5_OUT_HIGH)

// RTS
#define RTS_ASSERT()	(GPIOB->ODR &= GPIO4_OUT_MASK)
#define RTS_DEASSERT()	(GPIOB->ODR |= GPIO4_OUT_HIGH)

// PWRKEY (we're not really using this, it's always held LOW)
#define PWRKEY_ASSERT()		(GPIOA->ODR &= ~GPIO_ODR_8)
#define PWRKEY_DEASSERT()	(GPIOA->ODR |= GPIO_ODR_8)

// Power gate MOSFET
#define POWER_UP()		(GPIOA->ODR &= ~GPIO_ODR_11)
#define POWER_DOWN()	(GPIOA->ODR |= GPIO_ODR_11)

static uint8_t readLine(void)
{
	static uint8_t idxLine;

	smsState_t smsState = sms_state();
	while(1)
	{
		char c;
		if(!uart_read(&c, 1))
			break;

		if(c == '\r') // end of line
		{
			lineBuffer[idxLine] = '\0';
			idxLine = 0;
			return 1;
		}
		else if(c == '\n') // ignore \n
			continue;

		if(idxLine < BUFFLEN_LINE - 1)
			lineBuffer[idxLine++] = c;
		//else
		//	GSM_DBG("BUFF OVF %c\n", c); // This might cause the WWDG to timeout
		
		if(smsState == SMS_STATE_PROMT && c == '>')
		{
			lineBuffer[idxLine] = '\0';
			idxLine = 0;
			return 1;
		}
	}

	return 0;
}

static void unsolicited(void)
{
	// +CSQN: <rssi>,<ber>
	// <ber> (bit error rate) is only for GPRS
	if(prefix("+CSQN: ", lineBuffer))
	{
		signal = atoi(lineBuffer + 7);
		GSM_DBG("SIG: %u\n", signal);
	}
	else if(prefix("+CREG: ", lineBuffer))
	{
		creg = atoi(lineBuffer + 7);
		GSM_DBG("REG: %u\n", creg);
	}
	else if(prefix("+CMTI: ", lineBuffer))
		sms_notification();
}

void gsm_init()
{
	// A2 = UART TX
	// A3 = UART RX
	// A8 = PWRKEY (Open-drain, no pull-up)
	// A11 = Power gate(Open-drain, no pull-up)
	// B4 = RTS (Open-drain, no pull-up)
	// B5 = DTR (Open-drain, no pull-up)
	
	// Don't enable pull-ups for open-drain outputs, use the 2.8V GSM module pull-ups instead
	
	GPIOA->OTYPER |= GPIO11_OTYPE_OD | GPIO8_OTYPE_OD;
	GPIOB->OTYPER |= GPIO5_OTYPE_OD | GPIO4_OTYPE_OD;
	
	// TODO
	//GPIOA->PUPDR |= GPIO_PUPD(11, DOWN);
	//GPIOA->PUPDR |= GPIO_PUPD(8, DOWN);
	//GPIOB->PUPDR |= GPIO_PUPD(5, DOWN);
	//GPIOB->PUPDR |= GPIO_PUPD(4, DOWN);
	
	POWER_DOWN();
	//PWRSAVE_ON();
	//GPIOA->ODR |= GPIO11_OUT_HIGH;// | GPIO8_OUT_HIGH;
	//GPIOB->ODR |= GPIO5_OUT_HIGH | GPIO4_OUT_HIGH;

	GPIOA->MODER |= GPIO11_MODE_OUTPUT | GPIO8_MODE_OUTPUT;
	GPIOB->MODER |= GPIO5_MODE_OUTPUT | GPIO4_MODE_OUTPUT;

	PWRKEY_ASSERT();
	PWRSAVE_OFF();
	RTS_ASSERT();
	state = GSM_STATE_OFF;
}

static uint8_t command(char* strCommand, millis_t timeout, gsmcmd_callback_t callback)
{
	if(cmd.state != GSM_CMD_IDLE)
		return 0;

	if(strCommand != NULL)
		strcpy(cmd.command, strCommand);
	cmd.timeout = timeout;
	cmd.callback = callback;
	
	GSM_DBG("CMD DELAY\n");

	cmd.time = millis();
	cmd.state = (powerSave != 0) ? GSM_CMD_SKIPDELAY : GSM_CMD_DELAY;

	GSM_DBG("PWR SAVE OFF\n");
	PWRSAVE_OFF();
	powerSave = millis();

	return 1;
}

static void callback_startup(char* line, gsmCmdResult_t result)
{
	UNUSED(line);

	if(result == GSM_CMDRESULT_OK)
	{
		// Go to next command if last one is OK
		startupSequence++;
		startupTimer = 0;
	}
	else if(result == GSM_CMDRESULT_ERROR) // Error, wait a bit then send again
	{
		// TODO retry counter

		startupTimer = millis();
	}
	else if(result == GSM_CMDRESULT_TIMEOUT) // Timeout, send again without wait
	{
		// TODO retry counter

		startupTimer = 0;
	}
}

static void shutdown(void)
{
	uart_gsm(0);
	POWER_DOWN();
	PWRKEY_ASSERT();
	PWRSAVE_OFF();
	RTS_ASSERT();
	state = GSM_STATE_OFF;
	lastTurnoffTime = millis();
}

static void reset(void)
{
	GSM_DBG("** RESET **\n");
	shutdown();
	gsm_on();
}

static void callback_shutdown(char* line, gsmCmdResult_t result)
{
	UNUSED(line);

	if(result != GSM_CMDRESULT_DATA)
	{
		GSM_DBG("** OFF **\n");
		shutdown();
	}
}

void gsm_update()
{
	if(state == GSM_STATE_OFF)
		return;
	else if(state == GSM_STATE_POWERDOWN)
	{
		if(command("AT+COPS=2\r", 10000, callback_shutdown)) // TODO use AT+CPOWD=1 instead?
		{
			state = GSM_STATE_POWERINGDOWN;
			GSM_DBG("** POWERING DOWN **\n");
		}
	}

	// Command response took too long
	if(cmd.state != GSM_CMD_IDLE && cmd.state != GSM_CMD_DELAY && millis() - cmd.time >= cmd.timeout)
	{
		cmd.state = GSM_CMD_IDLE;
		GSM_DBG("CMD TO: (%lu) (%s)\n", cmd.timeout, cmd.command);

		gsmcmd_callback_t tmp = cmd.callback;
		if(tmp != NULL)
		{
			cmd.callback = NULL;
			tmp(NULL, GSM_CMDRESULT_TIMEOUT);
		}

		powerSave = millis();

		if(startupSequence == STARTUP_SEQ_NONE) 
		{
			// If we're doing normal commands (not startup sequence commands) and there are too many timeouts then reset the module
			// TODO if the callback runs a GSM command then this timeout reset will loose it
			
			// TODO what if the AT+COPS=2 shutdown command timesout, we don't want to reset, just force shutdown

			timeoutCounter++;
			if(timeoutCounter >= CMD_TIMEOUT_RESET_COUNT)
				reset();
		}
		else if(startupSequence - STARTUP_SEQ_OFFSET == 0)
		{
			// If we're doing the first AT startup sequence command and it times out too many times then reset the module
			timeoutCounter++;
			if(timeoutCounter >= 25)
				reset();
		}
	}

	// Send command after delay
	if((cmd.state == GSM_CMD_DELAY && millis() - cmd.time >= CMD_DELAY) || cmd.state == GSM_CMD_SKIPDELAY)
	{
		GSM_DBG("CMD SEND\n");
		uart_send(cmd.command, strlen(cmd.command));
		cmd.state = GSM_CMD_SENDING;
		cmd.time = millis();
	}

	// Read any new data
	while(readLine())
	{
		GSM_DBG("!LINE: %s\n", lineBuffer);

		if(cmd.state == GSM_CMD_SENDING && prefix("AT", lineBuffer)) // Command sent, waiting for echo
			cmd.state = GSM_CMD_WAITING; // Got echo, waiting for cmd response
		else if(cmd.state == GSM_CMD_WAITING) // We've got some data while waiting for a command response
		{
			// NOTE: Unsolicited messages can come through while waiting for a command response!
			// This includes SMS "+CMT" messages, like after an AT+COPS=0 command and before the OK response!
			
			gsmCmdResult_t result;

			if(strcmp(lineBuffer, "OK") == 0)
				result = GSM_CMDRESULT_OK;
			else if(strcmp(lineBuffer, "ERROR") == 0)
				result = GSM_CMDRESULT_ERROR;
			else
				result = GSM_CMDRESULT_DATA;

			gsmcmd_callback_t tmp = cmd.callback;

			if(result != GSM_CMDRESULT_DATA)
			{
				cmd.callback = NULL; // Command finished, clear callback
				cmd.state = GSM_CMD_IDLE;
				powerSave = millis();
			}

			if(tmp != NULL)
				tmp(lineBuffer, result);
			
			if(result != GSM_CMDRESULT_DATA)
				timeoutCounter = 0;
		}
		else if(startupSequence == STARTUP_SEQ_SMSREADY)
		{
			if(strcmp(lineBuffer, "SMS Ready") == 0)
			{
				GSM_DBG("** SMSREADY **\n");
				startupSequence = STARTUP_SEQ_NONE;
				state = GSM_STATE_ON;
			}
		}

		if(lineBuffer[0] == '+')
			unsolicited();
	}

	if(startupSequence == STARTUP_SEQ_POWERON)
	{
		// Must allow a few seconds after power down before power up again
		if(millis() - lastTurnoffTime >= TURN_BACK_ON_DELAY)
		{
			GSM_DBG("** POWERON **\n");
			startupSequence = STARTUP_SEQ_CMDDELAY;
			startupTimer = millis();

			PWRSAVE_OFF();
			RTS_DEASSERT();
			PWRKEY_ASSERT();
			POWER_UP();
			uart_gsm(1);
		}
	}
	else if(startupSequence == STARTUP_SEQ_CMDDELAY)
	{
		// Allow some time after power up before sending commands
		if(millis() - startupTimer >= TURN_ON_CMD_DELAY)
		{
			GSM_DBG("** CMDDELAY **\n");
			startupSequence = STARTUP_SEQ_OFFSET;
			startupTimer = 0;
		}
	}
	else if(startupSequence == STARTUP_SEQ_SMSREADY)
	{
		if(millis() - startupTimer >= SMSREADY_TIMEOUT)
		{
			GSM_DBG("** SMSREADY TO **\n");
			startupSequence = STARTUP_SEQ_NONE;
			state = GSM_STATE_ON;
		}
	}
	else if(startupSequence >= STARTUP_SEQ_OFFSET)
	{
		if(startupTimer == 0 || millis() - startupTimer >= startupConfig[startupSequence - STARTUP_SEQ_OFFSET]->retryDelay)
		{
			startupTimer = 0;
			const cmd_startup_t* nextCmd = startupConfig[startupSequence - STARTUP_SEQ_OFFSET];
			if(nextCmd != NULL)
				command((char*)nextCmd->str, nextCmd->timeout, callback_startup);
			else
			{
				GSM_DBG("** CMD SEQ COMPLETE **\n");
				startupSequence = STARTUP_SEQ_SMSREADY;
				startupTimer = millis();
			}
		}
	}

	if(state == GSM_STATE_ON && cmd.state == GSM_CMD_IDLE && powerSave != 0 && millis() - powerSave >= PWR_SAVE_DELAY)
	{
		GSM_DBG("PWR SAVE ON\n");
		PWRSAVE_ON();
		powerSave = 0;
	}
}

void gsm_sendRaw(char* data, uint8_t len)
{
	if(state == GSM_STATE_ON)
		uart_send(data, len);
}

uint8_t gsm_command(char* strCommand, millis_t timeout, gsmcmd_callback_t callback)
{
	if(state == GSM_STATE_ON)
		return command(strCommand, timeout, callback);
	return 0;
}

void gsm_commandUpdateTimeout(millis_t timeout)
{
	cmd.timeout = timeout;
}

char* gsm_getBuffer()
{
	if(state == GSM_STATE_ON && cmd.state == GSM_CMD_IDLE)
		return cmd.command;
	return NULL;
}

void gsm_on()
{
	if(state == GSM_STATE_OFF)
	{
		GSM_DBG("** ON **\n");
		
		// Clear input buffer
		char c;
		while(uart_read(&c, 1));

		state = GSM_STATE_TURNINGON;
		cmd.state = GSM_CMD_IDLE;
		startupSequence = STARTUP_SEQ_POWERON;
		startupTimer = 0;
		timeoutCounter = 0;
		
		startupTime = millis();
	}
}

void gsm_off()
{
	if(state == GSM_STATE_ON)
	{
		state = GSM_STATE_POWERDOWN;
		GSM_DBG("** POWER DOWN **\n");
	}
}

gsm_state_t gsm_getState()
{
	gsm_state_t tmp = state;
	if(tmp != GSM_STATE_ON && tmp != GSM_STATE_OFF)
		tmp = GSM_STATE_CHANGING;
	return tmp;
}

uint8_t gsm_isConnected()
{
	return (state == GSM_STATE_ON && (creg == 1 || creg == 5));
}

uint8_t gsm_signal()
{
	if(state == GSM_STATE_ON)
		return signal;
	return 99;
}
