/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

// RTC gains approx 20 seconds every 15 minutes (will trigger at 14:40)
// When set to 12 hours it triggers after 11 hours and 38 minutes

#include "inc/common.h"

// This is mainly for debugging
// 0 = 12 hours (normal)
// 1 = 15 minutes
// 2 = 30 seconds
// 3 = 1 second
// 4 = 10 seconds
#define RTC_ALARM_INTERVAL	0

static volatile uint8_t triggered;
static volatile uint8_t alarmCount;

static void setNextAlarm(void)
{
	// Enable write access
	RTC->WPR = 0xCA;
	RTC->WPR = 0x53;

	RTC->CR &= ~RTC_CR_ALRAE; // Disable alarm A
	while (!(RTC->ISR & RTC_ISR_ALRAWF));

#if RTC_ALARM_INTERVAL == 0
	// Trigger every 12 hours
	if(!((RTC->ALRMAR>>16) & 0x3F)) // Is currently 0
		RTC->ALRMAR = RTC_ALRMAR_MSK4 | (1<<20) | (2<<16); // Set to 12 BCD
	else
		RTC->ALRMAR = RTC_ALRMAR_MSK4; // Set to 0
#elif RTC_ALARM_INTERVAL == 1
	// Trigger every 15 mins
	uint8_t mins = (RTC->ALRMAR & 0x7F00)>>8;
	mins = ((mins>>4) * 10) | (mins & 0x0F);
	mins += 15;
	mins %= 60;
	RTC->ALRMAR = RTC_ALRMAR_MSK4 | RTC_ALRMAR_MSK3 | (mins / 10)<<12 | (mins % 10)<<8;
#elif RTC_ALARM_INTERVAL == 2
	// Trigger every 30 secs
	if((RTC->ALRMAR & 0x0000003F) == 0)
		RTC->ALRMAR = RTC_ALRMAR_MSK4 | RTC_ALRMAR_MSK3 | RTC_ALRMAR_MSK2 | 3<<4;
	else
		RTC->ALRMAR = RTC_ALRMAR_MSK4 | RTC_ALRMAR_MSK3 | RTC_ALRMAR_MSK2;
#elif RTC_ALARM_INTERVAL == 3
	// Trigger every 1 sec
	RTC->ALRMAR = RTC_ALRMAR_MSK4 | RTC_ALRMAR_MSK3 | RTC_ALRMAR_MSK2 | RTC_ALRMAR_MSK1;
#elif RTC_ALARM_INTERVAL == 4
	// Trigger every 10 secs
	uint8_t secs10 = (RTC->ALRMAR & 0xFF) >> 4;
	secs10++;
	if(secs10 >= 6)
		secs10 = 0;
	RTC->ALRMAR = RTC_ALRMAR_MSK4 | RTC_ALRMAR_MSK3 | RTC_ALRMAR_MSK2 | secs10<<4;
#else
	#error "Invalid RTC alarm interval"
#endif

	RTC->CR = RTC_CR_ALRAIE | RTC_CR_ALRAE; // Enable alarm A and alarm interrupt

	// Disable write access
	RTC->WPR = 0xFE;
	RTC->WPR = 0x64;
}

void rtc_init()
{
	RCC->APB1ENR |= RCC_APB1ENR_PWREN;
	PWR->CR |= PWR_CR_DBP; // Allow RTC write access
	RCC->APB1ENR &= ~RCC_APB1ENR_PWREN;

	// Enable LSI
	RCC->CSR |= RCC_CSR_LSION;
	while(!(RCC->CSR & RCC_CSR_LSIRDY));

	RCC->BDCR |= RCC_BDCR_BDRST; // Reset RTC domain
	RCC->BDCR &= ~RCC_BDCR_BDRST; // Reset RTC domain
	RCC->BDCR |= RCC_BDCR_RTCEN|RCC_BDCR_RTCSEL_1; // Enable RTC clock and set to LSI

	// Enable write access
	RTC->WPR = 0xCA;
	RTC->WPR = 0x53;

	RTC->ISR |= RTC_ISR_INIT; // Init phase
	while(!(RTC->ISR & RTC_ISR_INITF));

	RTC->PRER = (127<<RTC_PRER_PREDIV_A_Pos)|(311<<RTC_PRER_PREDIV_S_Pos); // 40kHz/128 => 312 Hz, 312Hz/312 => 1Hz
	RTC->TR = 0;//RTC_TR_PM | 0; // Set time
	RTC->ISR &= ~RTC_ISR_INIT; // End init phase

	// Disable write access
	RTC->WPR = 0xFE;
	RTC->WPR = 0x64;

	setNextAlarm();

	// Enable interrupt stuff
	NVIC_SetPriority(RTC_IRQn, 3);
	NVIC_EnableIRQ(RTC_IRQn);
	EXTI->IMR |= EXTI_IMR_IM17;
	EXTI->RTSR |= EXTI_RTSR_RT17;
}

uint32_t rtc_get()
{
	return RTC->TR;
}

uint8_t rtc_alarm(uint8_t clear)
{
	// TODO atomic? only set evey ~12 hours anyway...
	uint8_t trig = triggered;
	if(clear)
		triggered = 0;
	return trig;
}

uint8_t rtc_alarmMonthly()
{
	uint8_t trig = (alarmCount >= 60); // 12 hours * 60 = 30 days
	if(trig)
		alarmCount = 0;
	return trig;
}

// https://electronics.stackexchange.com/questions/116745/stm32-rtc-alarm-not-configuring-for-the-second-time
void RTC_IRQHandler(void)
{
	if(EXTI->PR & EXTI_PR_PIF17)
	{
		EXTI->PR |= EXTI_PR_PIF17;
		RTC->ISR &= ~RTC_ISR_ALRAF;
		triggered = 1;
		alarmCount++;
		setNextAlarm();
		//setLED(0, LED_TOGGLE);
	}
}
