/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

#define PDU_EXTRA_STUFF 0

// https://web.archive.org/web/20070217095837/http://www.dreamfabric.com/sms/deliver_fo.html
// https://en.wikipedia.org/wiki/GSM_03.40

#define TP_RP_BIT	0x80 // Reply path. Parameter indicating that reply path exists. 
#define TP_UDHI_BIT	0x40 // User data header indicator. This bit is set to 1 if the User Data field starts with a header 
#define TP_SRI_BIT	0x10 // Status report indication. This bit is set to 1 if a status report is going to be returned to the SME
#define TP_MMS_BIT	0x04 // More messages to send. This bit is set to 0 if there are more messages to send

// Message type indicator. Bits no 1 and 0 are both set to 0 to indicate that this PDU is an SMS-DELIVER
#define TP_MTI_MASK					0x03
#define TP_MTI_SMS_DELIVER_REPORT	0x00
#define TP_MTI_SMS_DELIVER			0x00
#define TP_MTI_SMS_SUBMIT			0x01
#define TP_MTI_SMS_SUBMIT_REPORT	0x01
#define TP_MTI_SMS_COMMAND			0x02
#define TP_MTI_SMS_STATUS_REPORT	0x02
#define TP_MTI_SMS_RESERVED			0x03 // This seems to be used as a TP_MTI_SMS_DELIVER :/

//#define MAX_PDU_LEN ((160 * 2) + (9 * 2) + 16)

#define TIMEOUT_BUFFER	5000
#define TIMEOUT_PROMT	5000
#define TIMEOUT_SEND	60000
#define TIMEOUT_SMSLIST	5000
#define TIMEOUT_SMSREAD	5000
#define TIMEOUT_SMSDEL	5000

// NOTE: Time and date is when the SMS was sent
typedef struct {
	uint8_t smscLen;
	// No SMSC data stuff
	uint8_t deliver_fo;
	uint8_t senderLen;
	uint8_t typeOfAddress;
	char sender[SMS_NUMBER_LEN];
#if PDU_EXTRA_STUFF
	uint8_t protocolId;
	uint8_t dataCoding;
	uint8_t year;
	uint8_t month;
	uint8_t day;
	uint8_t hour;
	uint8_t min;
	uint8_t secs;
	uint8_t tz;
#endif
	uint8_t charCount;
} pdu_t;

static sms_t sms;
static uint8_t smsId;
static uint8_t newSMS;
static smssent_callback_t sentCallback;

void __attribute__((weak)) CB_SMS_RECEIVE(sms_t* sms)
{
	UNUSED(sms);
}

static void swapChars(char* chars)
{
	char tmp = chars[0];
	chars[0] = chars[1];
	chars[1] = tmp;
}

#if PDU_EXTRA_STUFF
static uint8_t octet2dec(char* octet)
{
	swapChars(octet);
	char tmp[3];
	tmp[0] = octet[0];
	tmp[1] = octet[1];
	tmp[3] = '\0';
	return atoi(tmp);
}
#endif

static void dec2hex(uint8_t val, char* hex)
{
	static char hexVals[] = "0123456789ABCDEF";
	hex[0] = hexVals[val >> 4];
	hex[1] = hexVals[val & 0x0F];
}

static void decode7bit(char* data, uint8_t count)
{
	// Deocde hex string containing 7-bit characters into 8-bit characters and store in the SMS message buffer
	// 'count' is the number of expected characters after decoding

	uint8_t shift = 0;
	uint8_t nextVal = 0;

	if(count > SMS_MESSAGE_LEN - 1)
		count = SMS_MESSAGE_LEN - 1;

	for(uint8_t c=0,i=0;c<count;c++,i++)
	{
		uint8_t val = hex2dec(&data[i * 2]);

		uint8_t tmp = val >> (7-shift);
		val <<= shift;
		val |= nextVal;
		val &= 0x7F;

		nextVal = tmp;
		
		sms.message[c] = val;
		
		shift++;
		if(shift == 7)
		{
			sms.message[++c] = nextVal;
			nextVal = 0;
			shift = 0;
		}
	}
	
	sms.message[count] = '\0';
}

static uint8_t processPDU(char* buff, uint8_t pduPart, pdu_t* pdu)
{
	uint8_t skip = 2;

	switch(pduPart)
	{
		case 0:
			pdu->smscLen = hex2dec(buff) * 2; // Octets (2 hex chars)
			skip = pdu->smscLen + 2; // Skip SMSC data
			// TODO should a break be here?
			__attribute__((fallthrough));
		case 1:
			pdu->deliver_fo = hex2dec(buff);

			// Don't deal with UDHI (user data header), SMS is probably really long
			// Don't deal with RP (reply path) stuff
			// TODO what happens if we get a REPORT message? 
			if(pdu->deliver_fo & (TP_RP_BIT | TP_UDHI_BIT))
			{
				skip = 255;
				SMS_DBG("WARNING: Unsupported PDU, ignoring (%u)\n", pdu->deliver_fo);
			}
			break;
		case 2:
			pdu->senderLen = hex2dec(buff);
			if(pdu->senderLen > SMS_NUMBER_LEN - 1)
				pdu->senderLen = SMS_NUMBER_LEN - 1;
			break;
		case 3:
			pdu->typeOfAddress = hex2dec(buff);
			if(pdu->typeOfAddress == 0x91)
				sms.phone.type = NUMBERTYPE_INTERNATIONAL;
			else if(pdu->typeOfAddress == 0x81)
				sms.phone.type = NUMBERTYPE_UNKNOWN;
			else
			{
				skip = 255;
				SMS_DBG("WARNING: Unknown type-of-address, ignoring (%u)\n", pdu->typeOfAddress);
			}
			break;
		case 4:
		{
			uint8_t evenLen = (pdu->senderLen + 1) & 0xFE;

			memcpy(pdu->sender, buff, evenLen);
			for(uint8_t i=0;i<evenLen;i+=2)
				swapChars(&pdu->sender[i]);

			uint8_t nullPos = evenLen;
			if(nullPos == SMS_NUMBER_LEN || pdu->sender[nullPos - 1] == 'F')
				nullPos--;

			pdu->sender[nullPos] = '\0';
			
			memcpy(sms.phone.number, pdu->sender, SMS_NUMBER_LEN);

			skip = evenLen;
		}
			break;
#if PDU_EXTRA_STUFF
		case 5:
			pdu->protocolId = hex2dec(buff);
			break;
		case 6:
			pdu->dataCoding = hex2dec(buff);
			break;
		case 7:
			pdu->year = octet2dec(buff);
			break;
		case 8:
			pdu->month = octet2dec(buff);
			break;
		case 9:
			pdu->day = octet2dec(buff);
			break;
		case 10:
			pdu->hour = octet2dec(buff);
			break;
		case 11:
			pdu->min = octet2dec(buff);
			break;
		case 12:
			pdu->secs = octet2dec(buff);
			break;
		case 13:
			pdu->tz = octet2dec(buff);
			break;
#endif
		case 14:
			pdu->charCount = hex2dec(buff); // Expected char count after 7-bit decoding
			break;
		case 15:
			decode7bit(buff, pdu->charCount);
			skip = 0;
			break;
		default:
			break;
	}
	
	return skip;
}

static void decode_pdu(char* line)
{
	// http://www.turbocontrol.com/CoolStuff/SMS%20messages%20and%20the%20PDU%20format.htm

	pdu_t pdu;
	memset(&pdu, 0, sizeof(pdu_t));
	
	uint8_t pduPart = 0;
	while(1)
	{
		uint8_t skip = processPDU(line, pduPart, &pdu);
		if(!skip)
			break;
		else if(skip == 255)
			return;
		line += skip;
		pduPart++;
	}

	SMS_DBG("SMSC len: %u\n", pdu.smscLen / 2);
	SMS_DBG("Sender len: %u\n", pdu.senderLen);
	SMS_DBG("Sender: %s\n", pdu.sender);
	SMS_DBG("Deliver: %u\n", pdu.deliver_fo);
#if PDU_EXTRA_STUFF
	SMS_DBG("Protocol ID: %u\n", pdu.protocolId);
	SMS_DBG("Data coding scheme: %u\n", pdu.dataCoding);
	SMS_DBG("Time/date: %u/%u/%u %u:%u:%u-%u\n", pdu.year, pdu.month, pdu.day, pdu.hour, pdu.min, pdu.secs, pdu.tz);
#endif
	SMS_DBG("Char count: %u\n", pdu.charCount);
	SMS_DBG("Encoded: %s\n", line);
	SMS_DBG("Msg: %s\n", sms.message);
}

static void encode7bit(char* in, uint8_t inLen)//, char* out)
{
	// Encode 8-bit character string into 7-bit character data and send
	
	// Full GSM alphabet not supported
	// http://www.developershome.com/sms/gsmAlphabet.asp
	// https://en.wikipedia.org/wiki/GSM_03.38

	uint8_t shift = 0;
	//uint8_t c = 0;

	for(uint8_t i=0;i<inLen;i++)
	{
		char val = in[i];
		char next = in[i + 1];

		next <<= 7-shift;
		val >>= shift;
		
		val = next | val;
		
		//out[c++] = val;
		char buff[2];
		dec2hex(val, buff);
		gsm_sendRaw(buff, 2);

		shift++;
		if(shift == 7)
		{
			shift = 0;
			i++;
		}
	}

	//return c;
}

static void runSentCallback(smsSendResult_t status)
{
	smssent_callback_t tmp = sentCallback;
	if(tmp != NULL)
	{
		sentCallback = NULL;
		tmp(status);
	}
}

static uint8_t calcPDULen(void)
{
	uint8_t len = 8;

	uint8_t addrLen = strlen(sms.phone.number);
	len += ((addrLen + 1) & 0xFE) / 2;

	uint8_t msgLen = strlen(sms.message);
	len += msgLen - (msgLen / 8);

	return len;
}

static void makeAndSendPDU(void)
{
	char buff[2];

	// Very basic PDU assembly

	uint8_t addrLen = strlen(sms.phone.number);
	uint8_t msgLen = strlen(sms.message);

	gsm_sendRaw("001100", strlen("001100")); // Send SMSC length (0), SMS-SUBMIT, TP-Message-Reference
	dec2hex(addrLen, buff);
	gsm_sendRaw(buff, 2); // Send number length
	dec2hex(sms.phone.type, buff);
	gsm_sendRaw(buff, 2); // Type-of-Address

	// Odd length numbers must be made even by adding an F to the end
	if(addrLen & 0x1)
		sms.phone.number[addrLen] = 'F'; // WARNING: string is no longer null-terminated

	// Swap number chars and send
	for(uint8_t i=0;i<addrLen;i+=2)
	{
		buff[0] = sms.phone.number[i + 1];
		buff[1] = sms.phone.number[i];
		gsm_sendRaw(buff, 2);
	}

	// Null-terminate again
	sms.phone.number[addrLen] = '\0';

	// Set TP-PID, TP-DCS and TP-Validity-Period
	// TP-Validity-Period: 0xAD = 7 days
	// http://forum.btframework.com/index.php?topic=3331.0
	gsm_sendRaw("0000AD", strlen("0000AD"));
	dec2hex(msgLen, buff);
	gsm_sendRaw(buff, 2); // Set data length (number of decoded characters in message)

	// Encode to 7bit and send
	encode7bit(sms.message, msgLen);
}

static void callback_smsSend(char* line, gsmCmdResult_t result)
{
	if(result == GSM_CMDRESULT_DATA)
	{
		if(sms.state == SMS_STATE_PROMT)
		{
			if(line[0] == '>')
			{
				gsm_commandUpdateTimeout(TIMEOUT_SEND);

				SMS_DBG("GOT PROMT, NOW WRITING\n");
				// Got the promt, now send the message
				//sms.state = SMS_WRITING;
				makeAndSendPDU();
#if ENABLE_SMS_SEND
				char c = 0x1A;
#else
				char c = 0x1B;
#endif
				gsm_sendRaw(&c, 1);
				sms.state = SMS_STATE_SENDING;
				SMS_DBG("WRITING FIN\n");
			}
		}
	}
	else // OK / ERROR  / TIMEOUT
	{
		if(sms.state == SMS_STATE_SENDING)
		{
			SMS_DBG("!SMS: %s\n", line);

			uint8_t status;
			if(result == GSM_CMDRESULT_OK)
			{
				status = SMS_SEND_SUCCESS;
				SMS_DBG("SMS send success\n");
			}
			else // ERROR  / TIMEOUT
			{
				status = SMS_SEND_FAILED;
				SMS_DBG("SMS send failed\n");
			}

			sms.state = SMS_STATE_IDLE;
			
			runSentCallback(status);
		}
		else // SMS_STATE_PROMT
		{
			// Probably timed out when waiting for the '>' promt, which only happens if something has gone very wrong (like the GSM module resetting)
			SMS_DBG("SMS UNEXPECTED STATE\n");
			sms.state = SMS_STATE_IDLE;
			runSentCallback(SMS_SEND_FAILED);
		}
	}
}

void sms_init()
{
	
}

void sms_notification()
{
	newSMS = 1;
}

sms_t* sms_getBuffer()
{
	if(sms.state == SMS_STATE_IDLE)
		return &sms;
	//SMS_DBG("Busy (buff)\n");
	return NULL;
}

smsState_t sms_state()
{
	return sms.state;
}

uint8_t sms_busy()
{
	return (sms.state != SMS_STATE_IDLE || newSMS);
}

uint8_t sms_send(phoneNumber_t* phone, smssent_callback_t callback)
{
	if(sms.state != SMS_STATE_IDLE)
	{
		SMS_DBG("SMS Busy\n");
		return 0;
	}
	else if(gsm_getState() != GSM_STATE_ON)
	{
		SMS_DBG("GSM Busy\n");
		return 0;
	}

	if(phone != NULL)
		memcpy(&sms.phone, phone, sizeof(phoneNumber_t));

	if(!strlen(sms.phone.number))
		return 0;

	sentCallback = callback;

	sms.state = SMS_STATE_SENDSETUP;
	sms.timer = millis();
	SMS_DBG("SMS SETUP\n");

	return 1;
}

static void callback_smsReceive(char* line, gsmCmdResult_t result)
{
	//SMS_DBG("RX CB: %u %s\n", result, line);

	if(result == GSM_CMDRESULT_OK)
	{
		if(sms.state == SMS_STATE_LISTING)
		{
			if(smsId != 0)
				sms.state = SMS_STATE_READ;
			else
			{
				newSMS = 0;
				sms.state = SMS_STATE_IDLE;
				SMS_DBG("No SMSs to list\n");
			}
		}
		else if(sms.state == SMS_STATE_READING || sms.state == SMS_STATE_GETPDU || sms.state == SMS_STATE_GOTPDU)
			sms.state = SMS_STATE_DELETE;
		else if(sms.state == SMS_STATE_DELETEING)
			sms.state = SMS_STATE_DOCALLBACK;
	}
	else if(result == GSM_CMDRESULT_DATA)
	{
		if(line[0] == '+')
		{
			if(sms.state == SMS_STATE_LISTING && smsId == 0 && prefix("+CMGL: ", line)) // Find first unread SMS
				smsId = atoi(line + 7);
			else if(sms.state == SMS_STATE_READING && prefix("+CMGR: ", line)) // Next line will be the PDU
				sms.state = SMS_STATE_GETPDU;
		}
		else if(sms.state == SMS_STATE_GETPDU)
		{
			sms.state = SMS_STATE_GOTPDU;
			decode_pdu(line);
		}
	}
	else // ERROR / TIMEOUT
	{
		// If sms.state == SMS_STATE_LISTING then probably no SIM card inserted
		newSMS = 0;
		sms.state = SMS_STATE_IDLE;
		SMS_DBG("RX CB ERR/TO\n");
	}
}

void sms_update(uint8_t processNewSMS)
{
	if(sms.state == SMS_STATE_SENDSETUP)
	{
		char* cmdBuff = gsm_getBuffer();
		if(cmdBuff != NULL)
		{
			SMS_DBG("To: (%u - %s)\n", sms.phone.type, sms.phone.number);
			SMS_DBG("SENDING CMGS\n");
			sprintf(cmdBuff, "AT+CMGS=%u\r", calcPDULen());
			gsm_command(NULL, TIMEOUT_PROMT, callback_smsSend);
			sms.state = SMS_STATE_PROMT;
			SMS_DBG("WAITING FOR PROMT\n");
		}
		else if(millis() - sms.timer >= TIMEOUT_BUFFER)
		{
			SMS_DBG("SEND TIMEOUT, unable to get GSM CMD buffer\n");
			sms.state = SMS_STATE_IDLE;
			runSentCallback(SMS_SEND_TIMEOUT);
		}
	}

	// TODO reset GSM module if too many errors
	// TODO what if GSM is turned off in the middle of this?

	if(sms.state == SMS_STATE_DOCALLBACK)
	{
		sms.state = SMS_STATE_IDLE;
		CB_SMS_RECEIVE(&sms);
	}
	else if(newSMS && processNewSMS && sms.state == SMS_STATE_IDLE)
	{
		// Don't get new SMSs if we're still processing and sending the response to a previous SMS BEE command

		if(gsm_command("AT+CMGL=0,1\r", TIMEOUT_SMSLIST, callback_smsReceive)) // List all unread SMSs, but don't mark as read
		{
			smsId = 0;
			memset(&sms, 0, sizeof(sms_t));
			sms.state = SMS_STATE_LISTING;
		}
	}
	else if(sms.state == SMS_STATE_READ)
	{
		char* cmdBuff = gsm_getBuffer();
		if(cmdBuff != NULL)
		{
			sprintf(cmdBuff, "AT+CMGR=%u,0\r", smsId); // Read SMS and mark as read
			gsm_command(NULL, TIMEOUT_SMSREAD, callback_smsReceive);
			sms.state = SMS_STATE_READING;
		}
	}
	else if(sms.state == SMS_STATE_DELETE)
	{
		if(gsm_command("AT+CMGD=1,3\r", TIMEOUT_SMSDEL, callback_smsReceive)) // Delete all read SMSs
			sms.state = SMS_STATE_DELETEING;
	}
}
