/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

typedef struct{
	smsSend_t status;
	millis_t timer;
	millis_t delay;		// Delay time before retrying
	uint8_t retries;	// Retry counter
	uint8_t maxRetries;	// Max number of retries before failing
	uint8_t stop;		// Flag to stop the sending process
	millis_t lastSend;	// Time of last send attempt
	smsmgrmake_callback_t onMakeMessage; // Function to run when generating SMS message
	smsmgrcomplete_callback_t onComplete; // Function to run when SMS has been sent (or failed)
} smsSendCtrl_t;

static smsSendCtrl_t SMSSend[SMSMGRMSG_COUNT];
static smsMgrMsg_t currentActive;

static char* getSmsCtrlName(smsMgrMsg_t smsMgrMsg)
{
	static char* names[] = {
		"LOWBATT",
		"STOP",
		"STATUS",
		"TRACK",
		"CMDRESP",
		"STATUSMONTHLY"
	};
	if(smsMgrMsg >= (sizeof(names) / sizeof(char*)))
		return "UNKNOWN";
	return names[smsMgrMsg];
}

static void callback(smsSendResult_t status)
{
	smsSendCtrl_t* smsCtrl = &SMSSend[currentActive];
	if(status == SMS_SEND_SUCCESS)
	{
		smsCtrl->status = SMSSEND_SENT;
		MGR_DBG("(%s) SENT\n", getSmsCtrlName(currentActive));
	}
	else
	{
		smsCtrl->status = SMSSEND_DELAY;
		smsCtrl->timer = millis();
		MGR_DBG("(%s) FAILED (%u)\n", getSmsCtrlName(currentActive), smsCtrl->retries);
	}
}

void smsmgr_init()
{

}

void smsmgr_setup(smsMgrMsg_t smsMgrMsg, millis_t retryDelay, uint8_t maxRetries, smsmgrmake_callback_t onMakeMessage, smsmgrcomplete_callback_t onComplete)
{
	smsSendCtrl_t* smsCtrl = &SMSSend[smsMgrMsg];
	smsCtrl->delay = retryDelay;
	smsCtrl->maxRetries = maxRetries;
	smsCtrl->onMakeMessage = onMakeMessage;
	smsCtrl->onComplete = onComplete;
}

void smsmgr_update()
{
	if(gsm_getState() != GSM_STATE_ON)
		return;

	uint8_t ready[SMSMGRMSG_COUNT];

	for(uint8_t i=0;i<SMSMGRMSG_COUNT;i++)
	{
		smsSendCtrl_t* smsCtrl = &SMSSend[i];
		ready[i] = 0;
		if(smsCtrl->status != SMSSEND_IDLE && smsCtrl->status != SMSSEND_SENT && smsCtrl->stop)
		{
			smsCtrl->status = SMSSEND_IDLE;
			MGR_DBG("(%s) STOPPED\n", getSmsCtrlName(i));
			if(smsCtrl->onComplete)
				smsCtrl->onComplete(SMS_SEND_FAILED);
		}
		else if(smsCtrl->status == SMSSEND_DELAY)
		{
			if(millis() - smsCtrl->timer >= smsCtrl->delay)
			{
				smsCtrl->retries++;
				if(smsCtrl->retries >= smsCtrl->maxRetries)
				{
					smsCtrl->status = SMSSEND_IDLE;
					MGR_DBG("(%s) MAX RETRIES EXCEEDED\n", getSmsCtrlName(i));
					if(smsCtrl->onComplete)
						smsCtrl->onComplete(SMS_SEND_FAILED);
				}
				else
					smsCtrl->status = SMSSEND_DO;
			}
		}
		else if(smsCtrl->status == SMSSEND_DO && smsCtrl->onMakeMessage)
			ready[i] = 1;
		else if(smsCtrl->status == SMSSEND_SENT)
		{
			smsCtrl->status = SMSSEND_IDLE;
			if(smsCtrl->onComplete)
				smsCtrl->onComplete(SMS_SEND_SUCCESS);
		}
		// else SMSSEND_SENDING / SMSSEND_IDLE
	}

	sms_t* sms = sms_getBuffer();
	if(sms != NULL)
	{
		// Find SMS control that has been waiting the longest to send an SMS
		millis_t longestTimeAgo = 0;
		uint8_t oldestIdx = 255;
		millis_t now = millis();
		for(uint8_t i=0;i<SMSMGRMSG_COUNT;i++)
		{
			smsSendCtrl_t* smsCtrl = &SMSSend[i];
			if(ready[i])
			{
				millis_t lastSendTimeAgo = now - smsCtrl->lastSend;
				if(lastSendTimeAgo >= longestTimeAgo)
				{
					longestTimeAgo = lastSendTimeAgo;
					oldestIdx = i;
				}
			}
		}
		
		if(oldestIdx != 255)
		{
			smsSendCtrl_t* smsCtrl = &SMSSend[oldestIdx];
			
			phoneNumber_t* phone;
			smsCtrl->onMakeMessage(sms, &phone);

			currentActive = oldestIdx;
			smsCtrl->lastSend = millis();
			if(sms_send(phone, callback))
				smsCtrl->status = SMSSEND_SENDING;
			else
				callback(SMS_SEND_FAILED);
		}
	}
}

uint8_t smsmgr_begin(smsMgrMsg_t smsMgrMsg)
{
	smsSendCtrl_t* smsCtrl = &SMSSend[smsMgrMsg];
	if(smsCtrl->status != SMSSEND_IDLE)// || gsm_getState() != GSM_STATE_ON)
		return 0;

	MGR_DBG("(%s) BEGIN\n", getSmsCtrlName(smsMgrMsg));
	smsCtrl->status = SMSSEND_DO;
	smsCtrl->retries = 0;
	smsCtrl->stop = 0;
	return 1;
}

void smsmgr_stop(smsMgrMsg_t smsMgrMsg)
{
	smsSendCtrl_t* smsCtrl = &SMSSend[smsMgrMsg];
	if(smsCtrl->status == SMSSEND_IDLE || smsCtrl->status == SMSSEND_SENT)
		MGR_DBG("(%s) ALREADY STOPPED\n", getSmsCtrlName(smsMgrMsg));
	else
	{
		MGR_DBG("(%s) STOPPING...\n", getSmsCtrlName(smsMgrMsg));
		smsCtrl->stop = 1;
	}
}

smsSend_t smsmgr_status(smsMgrMsg_t smsMgrMsg)
{
	return SMSSend[smsMgrMsg].status;
}

uint8_t smsmgr_busy()
{
	for(uint8_t i=0;i<SMSMGRMSG_COUNT;i++)
	{
		if(SMSSend[i].status != SMSSEND_IDLE)
			return 1;
	}
	return 0;
}
