/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

#define BATT_HYSTERESIS 200UL
#define BATT_CHECK_INTERVAL 30000
#define BATT_LOW_COUNT 3
#define CMD_PREFIX_LEN 4
#define SMS_MISSED_CHECK_INTERVAL 60000
#define EEP_CHECK 0xAA
#define STARTUP_TRIG_DELAY 60000

typedef uint8_t (*bee_func_t)(char* param, sms_t* sms);

typedef enum{
	SMSCMDSTATUS_ERROR = 0,
	SMSCMDSTATUS_OK
} smsCmdStatus_t;

typedef enum{
	BATTSTATUS_OK = 0,
	BATTSTATUS_SENDING,
	BATTSTATUS_SENT
} battStatus_t;

typedef struct{
	uint8_t eepCheck;
	uint16_t trackInterval;
	//uint32_t wakeInterval;
	uint16_t wakeDuration;
	uint16_t trigIdleTime;
	uint16_t lowBattThres;
	phoneNumber_t phone;
	uint8_t sensitivity;
} config_t;

typedef struct{
	uint8_t enable; // Enable triggers
	uint8_t on; // Triggered
	uint8_t override; // Manual override from BEE+TRIGGER
	millis_t time;
} track_t;

typedef struct{
	char* param;
	bee_func_t func;
} bee_cmd_t;

typedef struct{
	char command[10];
	smsCmdStatus_t status;
	phoneNumber_t phone;
} cmd_resp_t;

static track_t track;
__attribute__((used, __section__(".EEPROM"))) static volatile const config_t eep_config = {0,0,0,0,0,{0, ""},0};
static config_t config;
static gps_t lastLockedData; // TODO clear when trigger is cleared
static uint8_t shakeCounter;
static uint8_t smsCheck;
static millis_t smsCheckStart;
static millis_t lastTrackingSMSTime;

static cmd_resp_t lastCmdResponse;
static battStatus_t batteryStatus;
static phoneNumber_t sendStatusSMSNumber;

extern char* fwBuild;

static void saveConfig(void)
{
	eeprom_emu_update((void*)&eep_config, &config, sizeof(eep_config));
}

static void loadConfig(void)
{
	memcpy(&config, (void*)&eep_config, sizeof(config));
	if(config.eepCheck != EEP_CHECK) // TODO
	{
		TRK_DBG("CONFIG EEPROM ERROR\n\n");

		config.eepCheck = EEP_CHECK;
		config.trackInterval = 180;
		//config.wakeInterval = 43200;
		config.wakeDuration = 300;
		config.trigIdleTime = 360;
		config.lowBattThres = 3650;
		config.phone.type = NUMBERTYPE_UNKNOWN;
		config.sensitivity = 0;
		strcpy(config.phone.number, "");

		saveConfig();
	}
	TRK_DBG("CONFIG\n");
	TRK_DBG("Number: (Type: %u) \"%s\"\n", config.phone.type, config.phone.number);
	TRK_DBG("Track interval: %u\n", config.trackInterval);
	//TRK_DBG("Wake interval: %lu\n", config.wakeInterval);
	TRK_DBG("Wake duration: %u\n", config.wakeDuration);
	TRK_DBG("Trig idle: %u\n", config.trigIdleTime);
	TRK_DBG("Low batt thres: %u\n", config.lowBattThres);
	TRK_DBG("Sensitivity: %u\n\n\n", config.sensitivity + 1);
}

static void makeTimeFromGPS(char* buff, gps_t* gps)
{
	memcpy(buff, gps->time, 2); // Hour
	memcpy(buff + 3, gps->time + 2, 2); // Mins
	memcpy(buff + 6, gps->time + 4, 2); // Secs
	memcpy(buff + 9, gps->date + 4, 2); // Year
	memcpy(buff + 12, gps->date + 2, 2); // Month
	memcpy(buff + 15, gps->date, 2); // Date
	buff[2] = ':';
	buff[5] = ':';
	buff[8] = ' ';
	buff[11] = '-';
	buff[14] = '-';
	buff[17] = '\0';
}

static void smsmgr_complete_track(smsSendResult_t status)
{
	UNUSED(status);
	lastTrackingSMSTime = millis();
}

static void smsmgr_complete_lowBatt(smsSendResult_t status)
{
	if(status == SMS_SEND_SUCCESS)
	{
		if(batteryStatus == BATTSTATUS_SENDING)
			batteryStatus = BATTSTATUS_SENT;
	}
	else
		batteryStatus = BATTSTATUS_OK; // Will send again
}

static uint16_t dischargeCurve[] = {
	4150, // 100%
	4020, // 90%
	3950,
	3870,
	3810,
	3780,
	3740,
	3700,
	3660,
	3600, // 10%
	3550, // 0%
};

#define DISCHARGE_CURVE_POINTS (sizeof(dischargeCurve) / sizeof(uint16_t))

static void make_status_sms(char* buff)
{
	battery_updateNow();

	uint32_t battVoltage = battery_voltage();
	uint8_t battPercentage = 0;
	
	for(uint8_t i=0;i<DISCHARGE_CURVE_POINTS;i++)
	{
		if(battVoltage >= dischargeCurve[i])
		{
			if(i == 0)
				battPercentage = 100;
			else
			{
				uint8_t tmp = dischargeCurve[i - 1] - dischargeCurve[i];
				uint8_t tmp2 = battVoltage - dischargeCurve[i];
				battPercentage = (100 - (i * (100 / (DISCHARGE_CURVE_POINTS - 1)))) + ((tmp2 * 10) / tmp);
			}
			break;
		}
	}

	char timedate[18];
	makeTimeFromGPS(timedate, &lastLockedData);

	gps_t* gpsInfo = gps_get(NULL);
	char* gpsStatus;
	if(gpsInfo->state == GPS_STATE_LOCKING)
		gpsStatus = "NO LOCK";
	else if(gpsInfo->state == GPS_STATE_LOCKED)
		gpsStatus = "LOCKED";
	else
		gpsStatus = "OFF";

	sprintf(
		buff,
		"BAT: %u%% %lumV (%umV)\nTMP: %ldC\nTRK: %c (%c)\nSIG: %u/31\nGPS: %s (%s%d.%lu,%s%d.%lu - %u KPH - %s)\nNUM: %s\nSEN: %u",
		battPercentage,
		battVoltage,
		config.lowBattThres,
		adc_temperature(),
		(track.on || track.override) ? 'Y' : 'N',
		track.enable ? 'Y' : 'N',
		gsm_signal(),
		gpsStatus,
		lastLockedData.deg_lat_isNegZero ? "-" : "",
		lastLockedData.deg_lat, // TODO dont clear last locked data?
		lastLockedData.deg_lat_frac,
		lastLockedData.deg_lon_isNegZero ? "-" : "",
		lastLockedData.deg_lon,
		lastLockedData.deg_lon_frac,
		lastLockedData.speed,
		timedate,
		(config.phone.number[0] != '\0') ? config.phone.number : "NOT SET",
		config.sensitivity + 1
	);

	TRK_DBG("\n----\n%s\n----\n", buff);
}

static void smsmgr_make_status(sms_t* sms, phoneNumber_t** phone)
{
	make_status_sms(sms->message);
	*phone = &sendStatusSMSNumber;
}

static void smsmgr_make_statusmonthly(sms_t* sms, phoneNumber_t** phone)
{
	make_status_sms(sms->message);
	*phone = &config.phone;
}

static void smsmgr_make_cmdResponse(sms_t* sms, phoneNumber_t** phone)
{
	sprintf(
		sms->message,
		"%s : %s",
		lastCmdResponse.command,
		(lastCmdResponse.status == SMSCMDSTATUS_OK) ? "OK" : "ERROR"
	);

	*phone = &lastCmdResponse.phone;

	TRK_DBG("\n----\n%s\n----\n", sms->message);
}

static void smsmgr_make_lowBatt(sms_t* sms, phoneNumber_t** phone)
{
	uint32_t voltage = battery_voltage();
	sprintf(
		sms->message,
		"LOW BATTERY: %lumV (%umV)",
		voltage,
		config.lowBattThres
	);
	*phone = &config.phone;
}

static void make_track_sms(char* buff, char* state)
{
	char timedate[18];
	makeTimeFromGPS(timedate, &lastLockedData);

	sprintf(
		buff,
		"TRK: (%s) https://maps.google.com/maps?q=loc:%s%d.%lu,%s%d.%lu - %u KPH - %s",
		state,
		lastLockedData.deg_lat_isNegZero ? "-" : "",
		lastLockedData.deg_lat,
		lastLockedData.deg_lat_frac,
		lastLockedData.deg_lon_isNegZero ? "-" : "",
		lastLockedData.deg_lon,
		lastLockedData.deg_lon_frac,
		lastLockedData.speed,
		timedate
	);
}

static void smsmgr_make_stop(sms_t* sms, phoneNumber_t** phone)
{
	make_track_sms(sms->message, "STOP");
	*phone = &config.phone;
	TRK_DBG("%s\n", sms->message);
}

static void smsmgr_make_track(sms_t* sms, phoneNumber_t** phone)
{
	gps_t* gpsInfo = gps_get(NULL); // TODO
	make_track_sms(sms->message, (gpsInfo->state == GPS_STATE_LOCKED) ? "LOCKED" : "NO LOCK");
	*phone = &config.phone;
	TRK_DBG("%s\n", sms->message);
}

void tracker_init()
{
	// GPIOs are used by everything, so enable them all here
	RCC->AHBENR |= RCC_AHBENR_GPIOAEN | RCC_AHBENR_GPIOBEN | RCC_AHBENR_GPIOFEN;

	// Pullups on unused pins
	GPIOA->PUPDR |= GPIO7_PUPD_UP | GPIO6_PUPD_UP | GPIO0_PUPD_UP;
	GPIOB->PUPDR |= GPIO0_PUPD_UP;

	// LEDs
	GPIOF->MODER |= GPIO1_MODE_OUTPUT | GPIO0_MODE_OUTPUT;

	eeprom_emu_init();
	millis_init();
	sleep_init();
	uart_init();
	uartsoft_init();
	i2c_init(); // TODO this has delays
	adc_init();
	rtc_init();
	gsm_init();
	gps_init();
	accel_init();
	battery_init();
	sms_init();
	smsmgr_init();

	for(uint8_t i=0;i<8;i++)
	{
		setLED(0, LED_ON);
		setLED(1, LED_OFF);
		delay(100);
		setLED(0, LED_OFF);
		setLED(1, LED_ON);
		delay(100);
	}

	setLED(0, LED_OFF);
	//setLED(1, LED_OFF);
	led_on(1);

	TRK_DBG("\n\n\nSTART\nSafeBee Tracker\nhttp://zakkemble.net\nFW: " FW_VERSION " (Built: %s)\n\n\n", fwBuild);

	smsmgr_setup(SMSMGRMSG_LOWBATT, 2000, 5, smsmgr_make_lowBatt, smsmgr_complete_lowBatt);
	smsmgr_setup(SMSMGRMSG_STOPPED, 500, 5, smsmgr_make_stop, NULL);
	smsmgr_setup(SMSMGRMSG_STATUS, 500, 5, smsmgr_make_status, NULL);
	smsmgr_setup(SMSMGRMSG_TRACK, 500, 1, smsmgr_make_track, smsmgr_complete_track);
	smsmgr_setup(SMSMGRMSG_CMDRESP, 500, 2, smsmgr_make_cmdResponse, NULL);
	smsmgr_setup(SMSMGRMSG_STATUSMONTHLY, 500, 5, smsmgr_make_statusmonthly, NULL);

	track.enable = 1;
	smsCheck = 1;
	smsCheckStart = millis();

	strcpy(lastLockedData.time, "000000");
	strcpy(lastLockedData.date, "000000");

	loadConfig();

	smsmgr_begin(SMSMGRMSG_STATUSMONTHLY);

	//track.override = 1; // TODO
}

uint8_t CB_SLEEP_OK(void)
{
	// NOTE: Try to keep this as short as possible. If an interrupt fires here (like the millisecond counter) then the MCU will immediately wake up and loop again.
	// It's not really much of an issue though.

	if(
//		uart_available() == 0 &&
		!accel_shake(0) &&
		!rtc_alarm(0) &&
		gps_get(NULL)->state == GPS_STATE_OFF &&
		gsm_getState() == GSM_STATE_OFF
	)
	{
		led_sleep();
		//TRK_DBG("SLEEP\n");
		return 1;
	}

	return 0;
}

static smsCmdStatus_t bee_status(char* param, sms_t* sms)
{
	UNUSED(param);
	if(smsmgr_begin(SMSMGRMSG_STATUS))
	{
		memcpy(&sendStatusSMSNumber, &sms->phone, sizeof(phoneNumber_t));
		return SMSCMDSTATUS_OK;
	}
	return SMSCMDSTATUS_ERROR;
}

static smsCmdStatus_t bee_trigger(char* param, sms_t* sms)
{
	UNUSED(param);
	UNUSED(sms);
	track.override = 1;
	track.time = millis();
	lastTrackingSMSTime = millis() - (config.trackInterval * 1000);
	return SMSCMDSTATUS_OK;
}

static smsCmdStatus_t bee_clear(char* param, sms_t* sms)
{
	UNUSED(param);
	UNUSED(sms);
	track.on = 0;
	track.override = 0;
	shakeCounter = 0;
	return SMSCMDSTATUS_OK;
}

static smsCmdStatus_t bee_disable(char* param, sms_t* sms)
{
	UNUSED(param);
	UNUSED(sms);
	track.enable = 0;
	return SMSCMDSTATUS_OK;
}

static smsCmdStatus_t bee_enable(char* param, sms_t* sms)
{
	UNUSED(param);
	UNUSED(sms);
	track.enable = 1;
	return SMSCMDSTATUS_OK;
}

static smsCmdStatus_t bee_number(char* param, sms_t* sms)
{
	UNUSED(sms);

	// Phone number to send tracking texts to

	if(param == NULL)
		return SMSCMDSTATUS_ERROR;

	// Check first character for a '+'
	numberType_t type = (*param == '+') ? NUMBERTYPE_INTERNATIONAL : NUMBERTYPE_UNKNOWN;

	// Go to next character if its a '+'
	if(type == NUMBERTYPE_INTERNATIONAL)
		param++;

	char* number = param;
	uint8_t len = 0;

	// Validate number - must only be digits
	for(;*param;param++)
	{
		len++;
		if(len > (SMS_NUMBER_LEN - 1) || *param < '0' || *param > '9')
			return SMSCMDSTATUS_ERROR;
	}

	memcpy(config.phone.number, number, SMS_NUMBER_LEN - 1);
	config.phone.type = type;
	config.phone.number[len] = '\0';
	saveConfig();
	TRK_DBG("New num: (Type: %u) %s\n", config.phone.type, config.phone.number);

	return SMSCMDSTATUS_OK;
}

static smsCmdStatus_t bee_sense(char* param, sms_t* sms)
{
	UNUSED(sms);

	if(param == NULL)
		return SMSCMDSTATUS_ERROR;
	
	uint8_t sense = (*param) - '1';
	if(sense > 2)
		return SMSCMDSTATUS_ERROR;

	config.sensitivity = sense;
	shakeCounter = 0;
	saveConfig();
	return SMSCMDSTATUS_OK;
}

static const bee_cmd_t bee_cmd[] = {
	{"STATUS",	bee_status},
	{"TRIGGER",	bee_trigger},
	{"CLEAR",	bee_clear},
	{"DISABLE",	bee_disable},
	{"ENABLE",	bee_enable},
	{"NUMBER",	bee_number},
	{"SENSE",	bee_sense}
};

#define BEE_CMD_COUNT (sizeof(bee_cmd) / sizeof(bee_cmd_t))

void CB_SMS_RECEIVE(sms_t* sms)
{
	// Command format:
	// BEE+<command>=<param>
	//
	// Examples:
	// BEE+NUMBER=+123456789
	// BEE+STATUS

	uint8_t length = strlen(sms->message);

	TRK_DBG("SMS from (%u - %s) (Len %u):\n%s\n", sms->phone.type, sms->phone.number, length, sms->message);

	// Convert to uppercasse for case insensitivity
	for(uint8_t i=0;i<length;i++)
		sms->message[i] = toupper(sms->message[i]);

	// Make sure SMS is long enough and starts with "BEE+"
	if(length <= CMD_PREFIX_LEN || !prefix("BEE+", sms->message))
		return;

	// Split command and parameter
	char* command = sms->message + CMD_PREFIX_LEN;
	char* param = strchr(command, '=');
	if(param != NULL)
	{
		*param = '\0'; // This will effectively split the param from the command string
		param++; // Point to beginning of param

		// Skip spaces
		while(*param == ' ')
			param++;
	}

	TRK_DBG("\n***\nBEECMD: %s\n***\n\n", command);
	TRK_DBG("\n***\nPARAM: %s\n***\n\n", (param != NULL) ? param : "(NONE)");

	smsCmdStatus_t status = SMSCMDSTATUS_ERROR;

	// Find and run function based on command
	for(uint8_t i=0;i<BEE_CMD_COUNT;i++)
	{
		if(strcmp(command, bee_cmd[i].param) == 0)
		{
			status = bee_cmd[i].func(param, sms);
			break;
		}
	}

	if(smsmgr_begin(SMSMGRMSG_CMDRESP))
	{
		memcpy(&lastCmdResponse.phone, &sms->phone, sizeof(phoneNumber_t));
		strcpy(lastCmdResponse.command, command);
		lastCmdResponse.status = status;
	}
}

static void lowBattCheck(void)
{
	// TODO disable tracking if below 3.4V

	// While the battery is below the treshold voltage we keep trying to send an SMS until success.
	// No more low battery SMSs will be sent until the voltage is back over the threshold plus the hysteresis, then goes back below the threshold.

	static uint8_t lowBattCount;
	static millis_t lastBattCheck;

	if(millis() - lastBattCheck > BATT_CHECK_INTERVAL)
	{
		lastBattCheck = millis();

		uint32_t voltage = battery_voltage();
		if(voltage && voltage < config.lowBattThres)
		{
			if(lowBattCount < BATT_LOW_COUNT)
			{
				lowBattCount++;
				TRK_DBG("LOWBATT D: %lumV %u\n", voltage, lowBattCount);
			}
		}
		else if(
			lowBattCount > 0 &&
			(
				(
					batteryStatus == BATTSTATUS_OK &&
					voltage >= config.lowBattThres
				) ||
				(
					batteryStatus != BATTSTATUS_OK &&
					voltage > config.lowBattThres + BATT_HYSTERESIS
				)
			)
		)
		{
			lowBattCount = 0;
			TRK_DBG("LOWBATT R: %lumV %u\n", voltage, lowBattCount);
		}

		if(batteryStatus == BATTSTATUS_OK && lowBattCount >= BATT_LOW_COUNT)
		{
			if(smsmgr_begin(SMSMGRMSG_LOWBATT))
				batteryStatus = BATTSTATUS_SENDING;
		}
		else if(batteryStatus != BATTSTATUS_OK && lowBattCount < BATT_LOW_COUNT)
		{
			smsmgr_stop(SMSMGRMSG_LOWBATT);
			batteryStatus = BATTSTATUS_OK;
		}
	}
}

static void checkTrigger(void)
{
	static uint8_t lastSecs;
	static uint8_t lastMins;
	static uint8_t startupDelayDone;

	uint8_t minutes = TIME_MINUTES(rtc_get());
	if(minutes != lastMins)
	{
		TRK_DBG("CLR TRIG CNTR: TO\n");
		lastMins = minutes;
		shakeCounter = 0;
	}

	if(!startupDelayDone && millis() < STARTUP_TRIG_DELAY)
	{
		accel_shake(1);
		return;
	}
	startupDelayDone = 1;

	if(track.enable)
	{
		if(accel_shake(1))
		{
			uint8_t seconds = TIME_SECONDS(rtc_get());
			if(seconds != lastSecs)
			{
				lastSecs = seconds;
				shakeCounter++;

				TRK_DBG("TRIG SECS: %u\n", shakeCounter);

				if(shakeCounter >= (config.sensitivity + 1))
				{
					TRK_DBG("** TRACKER TRIGGERED **\n");
					if(!track.on) // Only reset SMS timer if we've just started tracking
						lastTrackingSMSTime = millis() - (config.trackInterval * 1000); // Send an SMS straight away
					track.on = 1;
					track.time = millis();
					shakeCounter = 0;
				}
			}
		}

		// Keep tracking while GPS speed is more than 10KPH
		gps_t* gpsInfo = gps_get(NULL);
		if(gpsInfo->state == GPS_STATE_LOCKED && gpsInfo->speed > 10)
			track.time = millis();

		// Stop tracking if we havn't been triggered for a while
		if(
			track.on &&
			!track.override &&
			config.trigIdleTime &&
			millis() - track.time >= (config.trigIdleTime * 1000)
		)
		{
			TRK_DBG("CLR TRIG: IDLE\n");
			track.on = 0;
			smsmgr_begin(SMSMGRMSG_STOPPED);
		}
	}
}

static void periodicSMSWake(void)
{
	if(rtc_alarm(1))
	{
		TRK_DBG("** SMS CHECK START **\n");
		smsCheck = 1;
		smsCheckStart = millis();
		//sms_notification();
	}

	if(smsCheck)
	{
		gsm_on();

		if(millis() - smsCheckStart >= (config.wakeDuration * 1000))
		{
			smsCheck = 0;
			TRK_DBG("** SMS CHECK STOP **\n");
		}
	}
}

static void tracking(void)
{
	if(!track.on && !track.override)
		return;

	gps_on();
	gsm_on();

	uint8_t isNewData = 0;
	gps_t* gpsInfo = gps_get(&isNewData);

	if(gpsInfo->state == GPS_STATE_LOCKED)
	{
		if(isNewData)
			memcpy(&lastLockedData, gpsInfo, sizeof(gps_t));
	}

	if(millis() - lastTrackingSMSTime >= (config.trackInterval * 1000))
		smsmgr_begin(SMSMGRMSG_TRACK);
}

void tracker_update()
{
	uint8_t uartOverflow = uart_buffOverflow();
	if(uartOverflow)
		TRK_DBG("UART OVF %u\n", uartOverflow);

	led_update();
	battery_update();
	accel_update();
	gps_update();
	sms_update(smsmgr_status(SMSMGRMSG_CMDRESP) == SMSSEND_IDLE);
	gsm_update();
	smsmgr_update();

	lowBattCheck();
	checkTrigger();
	periodicSMSWake();
	tracking();

	if(rtc_alarmMonthly())
		smsmgr_begin(SMSMGRMSG_STATUSMONTHLY);

	// Check for SMSs just incase a new SMS notification was missed for some reason
	static millis_t lastNotificationCheck;
	if(millis() - lastNotificationCheck >= SMS_MISSED_CHECK_INTERVAL)
	{
		lastNotificationCheck = millis();
		if(gsm_getState() == GSM_STATE_ON)
			sms_notification();
	}

/*
	BEE+CONFIG=<track interval>,<wake interval>,<wake duration>,<trig idle time>,<low batt thres>
	implement a wake interval in seconds
	what if smsCheck stops, and we havnt got a GPS lock yet?
	what happens with no SIM
	what happens when no signal
	use GPRS
	inplement GSM reset on too many timeouts
	critical errors (WWDG, GSM reset)
	status report should go to the num that sent BEE+STATUS - DONE
	instead of GSM signal number, put POOR/OK/GOOD
	what happens without MOSFET slowdown capacitors?
	look at AT+IFC
	send AT before GSM command to make sure GSM module is still alive
*/

	// Debugging TODO
	static millis_t lastStatusReport;
	if(millis() - lastStatusReport >= 10000)
	{
		lastStatusReport = millis();
		battery_updateNow();
		TRK_DBG("BATT: %lumV\n", battery_voltage());
		TRK_DBG("TEMP: %ldC\n", adc_temperature());
	}

	if(!track.on && !track.override)
	{
		gps_off();
		if(!smsCheck && !smsmgr_busy() && !sms_busy())
			gsm_off();
	}

	if(sleep_now())
	{
		led_on(1);
		TRK_DBG("WAKE (RTC: %u | ACCEL: %u)\n", rtc_alarm(0), accel_shake(0));
		// NOTE: If these are both 0 then it was likely that the millisecond timer interrupt fired during the CB_SLEEP_OK() function.
	}
}
