/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

#define BAUD		115200
#define RXBUFFSIZE	255

// If the second UART output for PRINTF stuff shares pins with SWD then we only want to change the alternate function of the pins between 0 (SWD) and 1 (UART).
// The mode should not change between general output and alternate function.
// This is mainly because the SWD programmer doesn't seem to do a hard reset before communicating with the MCU so the SWD pins alternative function might be set to UART instead of SWD and thus SWD communication will fail.
#define PRINTF_UART_IS_SWD	1

// SWDCLK uses a pull-down, which is shared with UART TX for PRINTF stuff.
// When the pin mode is switched from PRINTF mode to SWD mode the pulldown will be fighting with the pullup of the serial-USB UART RX pin, and likely end up with a voltage close to VCC/2, which is basically a floating input.
// This option is for disabling the SWDCLK pulldown resistor.
// The pulldown should be enabled if UART PRINTF stuff is not shared with this pin, or if this pin is not going to be connected to a UART RX pin. Ifthe pulldown is disabled and nothing is connected to the pin then it will end up floating.
// 0 = Disable pulldown (should only be used for debugging if a serial-USB converter is connected)
// 1 = Keep pulldown enabled (should be used any other time)
#define PRINTF_SWDCLK_PULLDOWN	1

static volatile uint8_t buff[RXBUFFSIZE];
static ringBuffer_t rxBuffer;
static uint8_t gsmEnabled;
static volatile uint8_t buffOverflow;

static void init(void)
{
	RCC->APB2ENR |= RCC_APB2ENR_USART1EN; // Enable clock
	RCC->APB2RSTR |= RCC_APB2RSTR_USART1RST; // Pulse reset
	RCC->APB2RSTR &= ~RCC_APB2RSTR_USART1RST; // Pulse reset
	USART1->CR1 &= ~USART_CR1_UE; // Disable

	USART1->BRR = SystemCoreClock / BAUD;
	USART1->CR1 = USART_CR1_RXNEIE | USART_CR1_TE | USART_CR1_RE | USART_CR1_UE;
}

#define GSM_TX 2
#define GSM_RX 3
#define PRINTF_TX 14
#define PRINTF_RX 15

void uart_init()
{
	// PA2 = GSM TX
	// PA3 = GSM RX
	// PA14 = PRINTF TX (shared with SWCLK)
	// PA15 = PRINTF RX (not used, PA3 RX needs to stay listening to the GSM module)

	//GPIOA->PUPDR |= GPIO_PUPD(GSM_RX, UP); // GSM RX -> Pullup // TODO dont pullup here, let GSM module do it

	GPIOA->PUPDR |= GPIO_PUPD(PRINTF_RX, UP); // PRINTF_RX -> Pullup
	GPIOA->ODR |= GPIO_OUT(PRINTF_TX, HIGH); // PRINTF_TX -> Output HIGH
	GPIOA->AFR[0] |= GPIO_AFL(GSM_TX, 1) | GPIO_AFL(GSM_RX, 1); // GSM_TX and GSM_RX -> AF 1 (UART)
	GPIOA->MODER |= GPIO_MODE(GSM_TX, OUTPUT) | GPIO_MODE(GSM_RX, ALT); // GSM_TX -> OUTPUT | GSM_RX -> ALT

#if !PRINTF_SWDCLK_PULLDOWN && ENABLE_PRINTF_UART
	GPIOA->PUPDR &= GPIO_PUPD(PRINTF_TX, MASK); // Disable pulldown for PRINTF_TX / SWDCLK
#endif

	uart_gsm(0);
	
	rxBuffer.data = buff;

	NVIC_SetPriority(USART1_IRQn, 1);
	NVIC_EnableIRQ(USART1_IRQn);

	init();
}

uint8_t uart_buffOverflow()
{
	// TODO make this atomic, doesn't really matter too much, only used for debugging
	uint8_t tmp = buffOverflow;
	buffOverflow = 0;
	return tmp;
}

#if ENABLE_PRINTF_UART
void uart_printf(uint8_t enable)
{
	if(enable)
	{
		if(gsmEnabled) // Disable GSM_TX (force to OUTPUT HIGH), so data only comes from PRINTF_TX
			GPIOA->MODER = (GPIOA->MODER & GPIO_MODE(GSM_TX, MASK)) | GPIO_MODE(GSM_TX, OUTPUT); // GSM_TX -> OUTPUT

#if PRINTF_UART_IS_SWD
		// SWDCLK is already in ALT mode, so just change the ALT FUNCTION from SWD to UART
		GPIOA->AFR[1] |= GPIO_AFH(PRINTF_TX, 1); // PRINTF_TX -> AF1 (UART)
#else
		// Change pin mode from OUTPUT to ALT
		GPIOA->MODER = (GPIOA->MODER & GPIO_MODE(PRINTF_TX, MASK)) | GPIO_MODE(PRINTF_TX, ALT); // PRINTF_TX -> ALT
#endif
	}
	else
	{
#if PRINTF_UART_IS_SWD
		// Change ALT mode from UART to SWD
		GPIOA->AFR[1] &= ~(1<<GPIO_AFRH_AFSEL14_Pos); // PRINTF_TX -> AF0 (SWD) (unset AF1)
#else
		// Change pin mode from ALT to OUTPUT
		GPIOA->MODER = (GPIOA->MODER & GPIO_MODE(PRINTF_TX, MASK)) | GPIO_MODE(PRINTF_TX, OUTPUT); // PRINTF_TX -> OUTPUT
#endif

		if(gsmEnabled) // If GSM_TX was enabled then re-enabled it
			GPIOA->MODER = (GPIOA->MODER & GPIO_MODE(GSM_TX, MASK)) | GPIO_MODE(GSM_TX, ALT); // GSM_TX -> ALT
	}
}
#endif

void uart_gsm(uint8_t enabled)
{
	// NOTE: If this function was called after setting uart_printf to 1 then uart_printf(1) will need to be called again

	gsmEnabled = enabled;

	if(enabled)
	{
		GPIOA->ODR |= GPIO_OUT(GSM_TX, HIGH); // GSM_TX -> Output HIGH
		GPIOA->PUPDR &= GPIO_PUPD(GSM_RX, MASK); // No pullup/pulldown for GSM_RX, use GSM module builtin pullup
		GPIOA->MODER = (GPIOA->MODER & GPIO_MODE(GSM_TX, MASK)) | GPIO_MODE(GSM_TX, ALT); // GSM_TX -> ALT
	}
	else
	{
		// GSM module will be unpowered, these lines are to stop backpowering though the UART interface
		GPIOA->ODR &= GPIO_OUT(GSM_TX, MASK); // GSM_TX -> Output LOW
		GPIOA->PUPDR |= GPIO_PUPD(GSM_RX, DOWN); // GSM_RX - > Pulldown
		GPIOA->MODER = (GPIOA->MODER & GPIO_MODE(GSM_TX, MASK)) | GPIO_MODE(GSM_TX, OUTPUT); // GSM_TX -> OUTPUT
	}

	uart_printf(0);
}

uint8_t uart_available()
{
	int32_t diff = rxBuffer.head - rxBuffer.tail;
	if(diff < 0)
		diff += RXBUFFSIZE;

	return diff;
}

void uart_send(void* buff, uint8_t length)
{
	for(uint8_t i=0;i<length;i++)
	{
		USART1->TDR = ((uint8_t*)buff)[i];
		while(!(USART1->ISR & USART_ISR_TXE));
	}
	
	while(!(USART1->ISR & USART_ISR_TC));
}

uint8_t uart_read(void* buff, uint8_t length)
{
	uint8_t available = uart_available();
	if(length > available)
		length = available;

	for(uint8_t i=0;i<length;i++)
	{
		((uint8_t*)buff)[i] = rxBuffer.data[rxBuffer.tail];
		if(++rxBuffer.tail >= RXBUFFSIZE)
			rxBuffer.tail = 0;
	}

	return length;
}

void USART1_IRQHandler(void)
{
	if(!(USART1->ISR & USART_ISR_RXNE))
		return;

	int32_t diff = rxBuffer.head - rxBuffer.tail;
	if(diff < 0)
		diff += RXBUFFSIZE;

	if(diff < RXBUFFSIZE - 1)
	{
		rxBuffer.data[rxBuffer.head] = USART1->RDR;
		if(++rxBuffer.head >= RXBUFFSIZE)
			rxBuffer.head = 0;
	}
	else // Buffer full
	{
		USART1->RDR;
		if(buffOverflow < 255)
			buffOverflow++;
	}
}
/*
int _write(int file, char* ptr, int len)
{
	if (file == STDOUT_FILENO || file == STDERR_FILENO)
	{
		uart_send(ptr, len);
		return len;
	}
	errno = EIO;
	return -1;
}
*/
