/*
 * Project: SafeBee GPS Tracker
 * Author: Zak Kemble, contact@zakkemble.net
 * Copyright: (C) 2019 by Zak Kemble
 * License: 
 * Web: 
 */

#include "inc/common.h"

#define BAUD		9600
#define RXBUFFSIZE	255
#define RX_PORT		A
#define RX_PIN 		12

static volatile uint8_t buff[RXBUFFSIZE];
static ringBuffer_t rxBuffer;
static volatile uint8_t buffOverflow; // Debugging
static volatile uint8_t bitCount;
static volatile uint8_t	data;

#define ARR_VAL ((8000000UL / BAUD) - 1) // SystemCoreClock

void uartsoft_init()
{
	RCC->APB1ENR |= RCC_APB1ENR_TIM14EN; // Enable clock
	RCC->APB1RSTR |= RCC_APB1RSTR_TIM14RST; // Pulse reset
	RCC->APB1RSTR &= ~RCC_APB1RSTR_TIM14RST; // Pulse reset

	TIM14->DIER |= TIM_DIER_UIE;
	
	RCC->APB1ENR &= ~RCC_APB1ENR_TIM14EN;

	rxBuffer.data = buff;

	// Enable interrupt stuff
	//NVIC_SetPriority(TIM14_IRQn, 0);
	NVIC_EnableIRQ(TIM14_IRQn);

	//NVIC_SetPriority(EXTI4_15_IRQn, 0);
	NVIC_EnableIRQ(EXTI4_15_IRQn);
	EXTI->FTSR |= EXTI_FTSR_FT12;

	GPIO_SET_PUPD(RX_PORT, RX_PIN, DOWN);
}

uint8_t uartsoft_available()
{
	int32_t diff = rxBuffer.head - rxBuffer.tail;
	if(diff < 0)
		diff += RXBUFFSIZE;

	return diff;
}

uint8_t uartsoft_read(void* buff, uint8_t length)
{
	uint8_t available = uartsoft_available();
	if(length > available)
		length = available;

	for(uint8_t i=0;i<length;i++)
	{
		((uint8_t*)buff)[i] = rxBuffer.data[rxBuffer.tail];
		if(++rxBuffer.tail >= RXBUFFSIZE)
			rxBuffer.tail = 0;
	}

	return length;
}

void uartsoft_disable()
{
	__disable_irq();
	// Disable interrupts
	EXTI->IMR &= ~EXTI_IMR_IM12;
	TIM14->CR1 &= ~TIM_CR1_CEN;

	// Make sure theres no pending interrupts
	EXTI->PR |= EXTI_PR_PIF12;
	TIM14->SR &= ~TIM_SR_UIF;

	RCC->APB1ENR &= ~RCC_APB1ENR_TIM14EN;
	__enable_irq();

	GPIO_SET_PUPD(RX_PORT, RX_PIN, DOWN);
}

void uartsoft_enable()
{
	GPIO_SET_PUPD(RX_PORT, RX_PIN, UP);
	EXTI->IMR |= EXTI_IMR_IM12;
}

void TIM14_IRQHandler(void)
{
	if(TIM14->SR & TIM_SR_UIF)
	{
		TIM14->ARR = ARR_VAL;
		TIM14->SR &= ~TIM_SR_UIF;

		// Read pin
		uint8_t bit = (GPIOA->IDR & GPIO_IDR_12) >> 5;
		data >>= 1;
		data |= bit;

		bitCount--;
		if(!bitCount)
		{
			// Got all bits

			// Disable timer and interrupts
			TIM14->CR1 &= ~TIM_CR1_CEN;
			RCC->APB1ENR &= ~RCC_APB1ENR_TIM14EN;

			// Enable pin interrupt
			EXTI->IMR |= EXTI_IMR_IM12;

			int32_t diff = rxBuffer.head - rxBuffer.tail;
			if(diff < 0)
				diff += RXBUFFSIZE;

			if(diff < RXBUFFSIZE - 1)
			{
				rxBuffer.data[rxBuffer.head] = data;
				if(++rxBuffer.head >= RXBUFFSIZE)
					rxBuffer.head = 0;
			}
			else // Buffer full
			{
				if(buffOverflow < 255)
					buffOverflow++;
			}
		}
	}
}

void EXTI4_15_IRQHandler(void)
{
	if(EXTI->PR & EXTI_PR_PIF12)
	{
		EXTI->IMR &= ~EXTI_IMR_IM12;
		EXTI->PR |= EXTI_PR_PIF12;

		// 8 data bits + start bit
		bitCount = 9;
		
		// Enable timer
		RCC->APB1ENR |= RCC_APB1ENR_TIM14EN;
		TIM14->ARR = ARR_VAL / 2;
		TIM14->CNT = 0;
		TIM14->CR1 |= TIM_CR1_CEN;
	}
}
